/*****************************************************************************
* libspy.c        sort block access routines
*                 These functions can be used by user programs
*                 to access data blocks held in shared memory by the tape server
*     This version Mar 2003  J.Cresswell and V.Pucknell
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#ifdef LINUX
#include <sys/types.h>
#include <sys/shm.h>
#endif

#ifdef SOLARIS2
#include <unistd.h>
#include <sys/mman.h>
#include <fcntl.h>
#endif

#ifdef WIN32
#include <windows.h>
#endif 

#include  <tsspy.h>
#include <libspy.h>

#define MAX_STREAM 1
#define MAX_BUFFER_SIZE 65536

#ifdef LINUX
   int shmid;
   key_t shmkey = 110205;
#endif

#ifdef SOLARIS2
   int shmid;
   int shmkey = 110205;
#endif

#ifdef WIN32
   HANDLE shmid;
   int shmkey = 110205;
#endif

#ifndef LINUX
char object_name[16];
#endif

void * shm_bufferarea;
BUFFER_HEADER * baseaddress;

int number_of_buffers;
int buffers_offset;
int next_index;
unsigned long long current_age;

static int debug=0;

#ifdef WIN32
void ReportError (char * s) {

  LPVOID lpMsgBuf;
  DWORD ErrNo;
  
       ErrNo = GetLastError();
       FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                      NULL,
                      ErrNo,
                      0,
                      (LPTSTR) &lpMsgBuf,
                      0,
                      NULL
                      );
                      
       printf("%s: error %d - %s\n", s, ErrNo, lpMsgBuf);
       LocalFree (lpMsgBuf);
}
#endif

void sortSpyDebug (int opt) {debug = opt;}

/*****************************************************************************
* sortSpyOpen     make a connection to an the shared file for that stream
*                 return OK or ERROR (0 or -1)
*/

int sortSpyOpen (int stream)
{

  if (stream < 1 || stream > MAX_STREAM)
    {
      perror ("sortSpyOpen - stream number out of range");
      return -1;
    }
  
#ifdef SOLARIS2

/*    create a file mapped object (MASTER) or obtain ID of existing object */

    sprintf(object_name,"/SHM_%d", shmkey);

    shmid = shm_open(object_name, O_RDONLY, (mode_t) 0);
    if (shmid == -1)  {
        perror("shm_open");
        exit(1);
    }

#endif

#ifdef LINUX

  shmid = shmget(shmkey, 0, SHM_R);
  if (shmid == -1) {perror("shmget"); return -1;}
  
#endif

#ifdef WIN32

/*    create a file mapped object */

    sprintf(object_name,"SHM_%d", shmkey);

    shmid = CreateFileMapping( (HANDLE) INVALID_HANDLE_VALUE, NULL,
                               PAGE_READONLY,
                               (DWORD) 0,
                               (DWORD) 1,
                               object_name);

    if (!shmid)  {

/*    give up! */

             ReportError("CreatFileMapping");
             ExitProcess(1);
    }
#endif

#ifdef LINUX

  shm_bufferarea = shmat(shmid, (void *) 0, SHM_RDONLY);
  if (shm_bufferarea == (void *) -1) {perror("shmat"); return -1;}

#endif

#ifdef SOLARIS2

/*    attach the memory segment */

    shm_bufferarea = mmap((caddr_t) NULL, (size_t) 0x400000, PROT_READ, MAP_SHARED, shmid, (off_t) 0);
    if (shm_bufferarea == (void *) MAP_FAILED) {
        perror("mmap");
        exit(1);
    }

    close(shmid);

#endif

#ifdef WIN32

/*    map a view of the file mapped object */

    shm_bufferarea = MapViewOfFile(shmid,
                                 FILE_MAP_READ,
                                 0, 0, 0);

    if (shm_bufferarea == NULL) {
    
        ReportError("MapViewOfFile");
        ExitProcess(1);
    }

#endif
  
  printf("TSspy Shared buffer area (shmid %d) located at 0x%x\n",
	 shmid,shm_bufferarea);
  
  baseaddress = (BUFFER_HEADER *) shm_bufferarea;

  number_of_buffers = baseaddress->buffer_number;

  buffers_offset = baseaddress->buffer_offset;

  /* Flush all old buffers by setting current index ... */

  next_index = baseaddress->buffer_next;
  current_age = baseaddress->buffer_currentage;

  printf("TSspy Current age %lld index %d\n", current_age,next_index);
  
  return 0;
}

/*****************************************************************************
* sortSpyClose       close a connection to an the shared file for that stream
*                 close the file
*                 return OK or ERROR
*/

int sortSpyClose (int stream)
{
  int ret;

  if (stream < 1 || stream > MAX_STREAM)
    {
      perror ("sortSpyClose - stream number out of range");
      return -1;
    }

#ifdef LINUX

    (void) shmdt(shm_bufferarea);

#endif

#ifdef SOLARIS2

/*    detach the memory segment */

    (void) munmap(shm_bufferarea, (size_t) 0x400000);

#endif

#ifdef WIN32

/*    map a view of the file mapped object */

    UnMapViewOfFile(shm_bufferarea);

#endif
  
  printf("TSspy Shared buffer area (shmid %d) detached\n",shmid);
  
  return 0;
}

/*****************************************************************************
* sortSpyRead        get a block of data for that field
*                    check that the stream is valid and opened
*                    check that there is any data in the queue
*                    increment the queue read index 
*                    return length of data block or ERROR
*/

int sortSpyReadWithSeq (int stream, char *data, int length, int *seq)
{
  int next,index;
  int * bufferaddress;
  int i,len;
  int *ptr;


    if (stream < 1 || stream > MAX_STREAM)
    {
      perror ("sortSpyRead - stream number out of range");
      return -1;
    }

    len = 0;

 retry:

    if(baseaddress->buffer_age[next_index] != 0)
    {
        if(baseaddress->buffer_age[next_index] >= current_age)
	{
	   current_age = baseaddress->buffer_age[next_index];
	  
	   len = baseaddress->buffer_length;
	   if (!len) len = MAX_BUFFER_SIZE;

	   bufferaddress = (int *) ((char *)shm_bufferarea + buffers_offset + (len * next_index));

	   if(length < len) len = length;

	   if(debug)
	      printf ("sortSpyRead stream %d: Age %lld Index %d Buffer length %d\n",
		    stream,current_age,next_index,len);

	   ptr = (int *)data;
	   for (i = 0; i < (len >> 2); i++)
	   {
	       ptr[i] = bufferaddress[i];
	   }
           *seq = (int) current_age;
	  
	   if (current_age != baseaddress->buffer_age[next_index])
	   {
	        if(debug)
	            printf ("sortSpyRead stream %d: Copied oldage %lld newage %lld\n", stream, current_age, baseaddress->buffer_age[next_index]);
	      
	       goto retry;
	   }

	   next_index = ++next_index & (number_of_buffers -1);
	}
    } 
    else
    {
       if(debug)
          printf ("sortSpyRead - stream %d has no data\n",stream);
    }

    if(debug)
       printf("sortSpyRead - Stream %d length %d\n",stream,len);
  
    return len;                      /* return actual length of data block */
}
/******************************************************************************
* sortSpyRead   get a block of data for that stream
* as for sortSpyReadWithSeq, but ignoring the sequence numbering
* provided for compatibility with earlier version
*/

int sortSpyRead (int stream, char *data, int length) {
   int seq;
   return sortSpyReadWithSeq (stream, data, length, &seq);
   }
/*****************************************************************************/
